/*
 *   (C) Tomas Simonaitis v0.1
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/time.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <libnet.h>
#include <pcap.h>
#include <pthread.h>

#define SLEEP_TIME 20
#define SEND_SLEEP 20
#define DEFAULT_REPEAT_SLEEP 100
#define DEFAULT_TIMEOUT 6000
#define MIN_TIMEOUT 3
#define MAX_IPS 10000
#define IP_ALEN 4

/* Ethernet Header */
struct ethernet_header {
	u_char ether_dhost[ETHER_ADDR_LEN];
	u_char ether_shost[ETHER_ADDR_LEN];
	u_short ether_type;
};

/* ARP packet */
struct arp_pckt {
	u_int16_t hw_space; // Hardware address space (e.g., Ethernet, Packet Radio Net.)
	u_int16_t pr_space; // 16.bit: (ar$pro) Protocol address space.  For Ethernet hardware, this is from the set of type fields ether_typ$<protocol>.
	u_int8_t  hw_len;   // 8.bit: (ar$hln) byte length of each hardware address
	u_int8_t  pr_len;   // 8.bit: (ar$pln) byte length of each protocol address
	u_int16_t op_code;  // 16.bit: (ar$op)  opcode (ares_op$REQUEST | ares_op$REPLY)

	u_char hw_sender[ETHER_ADDR_LEN];
	u_char pw_sender[4];	// doh, can't use in_addr (or u32int) for some reason (align?)
	u_char hw_target[ETHER_ADDR_LEN];
	u_char pw_target[4];	// ^^
};


struct ip_mac {
	u_int32_t ip;
//	u_char mac[ETHER_ADDR_LEN];
	int found;
};

/* array of ips/macs */
struct ip_mac *req_ips;
int req_count = 0;
int req_found = 0;
int accepting = 0;
u_char mymac[ETHER_ADDR_LEN];
u_char zeromac[ETHER_ADDR_LEN];
u_char bcmac[ETHER_ADDR_LEN];
u_int32_t srcip;
int timeout = DEFAULT_TIMEOUT;
int rsleep = DEFAULT_REPEAT_SLEEP;

static libnet_t *libnet;

// -------------- //
void print_mac(const u_char *mac)
{
	int i;
	for (i=0; i<ETHER_ADDR_LEN; i++)
		printf("%.2X%c", mac[i], (i<5)?':':' ');

}

void packet_handler(const struct pcap_pkthdr *header, const u_char *packet)
{
	const struct ethernet_header *ethernet;
	const struct arp_pckt *arp;
	struct in_addr *ip_sender;
	struct in_addr *ip_target;
	int i, k;

	ethernet = (struct ethernet_header *)(packet);

	arp = (struct arp_pckt *) (packet + sizeof(struct ethernet_header));

	ip_sender = (struct in_addr *) (arp->pw_sender);
	ip_target = (struct in_addr *) (arp->pw_target);

	if (strncmp(arp->hw_sender, ethernet->ether_shost, ETHER_ADDR_LEN) != 0) {
		printf("Warning: MAC in arp packet doesn't match one in ethernet header, ");
		print_mac(arp->hw_sender);
		print_mac(ethernet->ether_shost);
		printf("\n");
	}

	for (k=0; k<req_count; k++)
		if (!req_ips[k].found && req_ips[k].ip == ip_sender->s_addr) {
			print_mac(ethernet->ether_shost);
			printf("%s\n", inet_ntoa(*ip_sender));

			req_ips[k].found = 1;
			req_found++;
		}
}

void arp_ping(u_int32_t dstip)
{
	static libnet_ptag_t arp=0, eth=0;

	if ((arp = libnet_build_arp(ARPHRD_ETHER,
					ETHERTYPE_IP,
					ETHER_ADDR_LEN,
					IP_ALEN,
					ARPOP_REQUEST,
					mymac,
					(u_int8_t*)&srcip,
					zeromac,
					(u_int8_t*)&dstip,
					NULL,
					0,
					libnet,
					arp)) == -1) {

		fprintf(stderr, "Error - libnet_build_arp(): %s\n",
				libnet_geterror(libnet));
	}
	if ((eth = libnet_build_ethernet(bcmac,
					mymac,
					ETHERTYPE_ARP,
					NULL,
					0,
					libnet,
					eth)) == -1) {
		fprintf(stderr, "Error - libnet_build_ethernet(): %s\n",
				libnet_geterror(libnet));
	}

	if (libnet_write(libnet) == -1) {
		fprintf(stderr, "Error - libnet_write(): %s\n", 
				libnet_geterror(libnet));
	}

}


void* ping_thread()
{
	int i, first_loop = 1;
	
	// send questions, repeat if we got time
	while (first_loop || (timeout > 0 && req_count > req_found && rsleep > 0) ) {
		for (i=0; i<req_count; i++) 
			if (!req_ips[i].found) {
				arp_ping(req_ips[i].ip);
				if (timeout <= 0) break;
				usleep(SEND_SLEEP);
			}

		if (!first_loop && rsleep > 0) usleep(rsleep);		
		first_loop = 0;
	}

}

void dhelp(char *pname)
{
	fprintf(stderr, "Usage: %s [-t timeout] [-i interface] [-r repeat wait] < ip1 ip2... | -f filename >\n", pname);
}

//
void add_ip(char *ip)
{
	struct in_addr tmp;	
	if (inet_aton(ip, &tmp)) {
		req_ips[req_count].found = 0;
		req_ips[req_count].ip = tmp.s_addr;
		req_count++;
	}
}

//
void add_ip_from_str(char *str)
{
	int i, k=0;
	char ip[16];

	for (i=0; i<strlen(str); i++) {
		if ( ((str[i] >= '0' && str[i] <= '9') || str[i] == '.') && k < 15 )
			ip[k++] = str[i]; else {
				ip[k] = '\0';
				if (k > 7) add_ip(ip);
				k = 0;
			}
	}

	ip[k] = '\0';
	if (k > 7) add_ip(ip);
}

// get ips from arguments
int get_argument_targets(int argc, char *argv[])
{
	// get ips list
	req_ips = (struct ip_mac *) malloc(argc * sizeof(struct ip_mac));

	struct in_addr tmp;
	int i;
	for (i=optind; i<argc; i++)
			add_ip(argv[i]);
	
	return (req_count != 0);
}

// get ips from file/stdin
int get_file_targets(char *filename)
{
	FILE *f;
	char t_str[256];
	
	if (strcmp(filename, "-") == 0) 
		f = stdin; else
			f = fopen(filename, "r");
	if (f == NULL)
		return 0;

	req_ips = (struct ip_mac *) malloc (MAX_IPS * sizeof(struct ip_mac));

	while (!feof(f) && req_count < MAX_IPS) {		
		if (fgets(t_str, 256, f) != NULL)
			add_ip_from_str(t_str);
	}

	return (req_count != 0);
}


int main(int argc, char *argv[])
{
	char *device = NULL;
	char *filename = NULL;
	char errbuf[PCAP_ERRBUF_SIZE];
	int pid;

	pcap_t *descriptor;
	struct pcap_pkthdr *header;
	u_char *packet;
	struct bpf_program fp;

	u_char *cp;
	int c;

	pthread_t thread;
	pthread_attr_t pthread_custom_attr;

	if (argc < 2) {
		dhelp(argv[0]);
		return 1;
	}

	if (getuid() && geteuid()) {
		fprintf(stderr, "Run marping as root\n");
		return 1;
	}

	while ( (c = getopt(argc, argv, "hi:t:r:f:")) != -1 )
		switch(c) {
			case 'h':
				dhelp(argv[0]);
				return 0;

			case 'i':
				if (optarg == NULL) {
					fprintf(stderr, "Specify interface name (e.g. eth0) after -i\n");
					return 1;
				} else
					device = optarg;
				break;
			case 't':
				if (optarg != NULL) {
					timeout = atoi(optarg);
					if (timeout < MIN_TIMEOUT) {
						fprintf(stderr, "Invalid timeout,  should be > %d\n", MIN_TIMEOUT);
						return 1;
					}
				}
				break;
			case 'r':
				if (optarg != NULL) {
					rsleep = atoi(optarg);
					if (rsleep < 0 || rsleep > timeout) {
						fprintf(stderr, "Invalid repeat wait, should be >= 0 and < %d\n", timeout);
						return 1;
					}
				}
				break;
			case 'f':
				if (optarg == NULL) {
					fprintf(stderr, "Specify file name or - for stdin\n");
					return 1;
				} else
					filename = optarg;
				break;
			default:
				dhelp(argv[0]);				
				return 1;
		}	
	//
	if (filename != NULL)
		c = get_file_targets(filename); else
			c = get_argument_targets(argc, argv);
	if (!c) {
		fprintf(stderr, "No ips found\n");
		dhelp(argv[0]);
		return 1;
	}

	//
	memset(bcmac, 0xff, sizeof(bcmac));
	memset(zeromac, 0, sizeof(zeromac));
	//

	if (device == NULL)
		if ((device = pcap_lookupdev(errbuf)) == NULL) {
			fprintf(stderr, "Error pcap_lookupdev\n");
			return 1;
		}


	if (!(libnet = libnet_init(LIBNET_LINK, device, errbuf)) )
	{ 
		fprintf(stderr, "Error - libnet_init: %s\n", errbuf); 
		return 1;
	}

	if (!(cp = (char*)libnet_get_hwaddr(libnet))) {
		fprintf(stderr, "Error - libnet_get_hwaddr(): %s\n",
				libnet_geterror(libnet));
		return 1;
	}

	memcpy(mymac, cp, ETHER_ADDR_LEN);

	if ( (srcip = libnet_get_ipaddr4(libnet)) == -1) {
		fprintf(stderr, "Error - libnet_get_ipaddr4(libnet): "
				"%s\n", libnet_geterror(libnet));
		return 1;
	}


	if((descriptor = pcap_open_live(device, BUFSIZ, 1, 0, errbuf)) == NULL) {
		fprintf(stderr, "Error - pcap_open_live\n");
		return 1;
	}

	if(pcap_compile(descriptor, &fp, "arp", 0, -1) == -1) {
		fprintf(stderr, "Error - pcap_compile\n");
		return 1;
	}

	if(pcap_setfilter(descriptor, &fp) == -1) {
		fprintf(stderr, "Error - pcap_setfilter\n");
		return 1;
	}

	pcap_freecode(&fp);

	if (pcap_setnonblock(descriptor, 1, errbuf) == -1) {
		fprintf(stderr, "Error - pcap_setnonblock\n");
		return 1;
	}

	pthread_attr_init(&pthread_custom_attr);
	pthread_create(&thread, &pthread_custom_attr, ping_thread, NULL);
	
	// wait for answers
	while (timeout > 0 && req_count > req_found) {
		while (pcap_next_ex(descriptor, &header, &packet) == 1)
			packet_handler(header, packet);
		usleep(SLEEP_TIME);
		timeout -= SLEEP_TIME;
	}

	pthread_join(thread, NULL);
	pcap_close(descriptor);

	return 0;
}



